{-| Module      : Gargantext.Core.Viz.Graph.ProxemyOptim
Description : Proxemy
Copyright   : (c) CNRS, 2017-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

(/!\ To be published soon, confidential document for now. After
publication it will be integrated to the backend with usual license --
see above)

Article: Confluence for Graph Clustering, B. Gaume and A. Delanoë, Journal:TBA

Code written in Haskell by A. Delanoë from first Python Specifications by B.
Gaume.

-}

{-# LANGUAGE FlexibleContexts
           , DataKinds
           , GADTs
           , KindSignatures
           , ScopedTypeVariables
           , StandaloneDeriving
           , TypeOperators
           , TypeApplications
           , NoImplicitPrelude
           , RankNTypes
           , MultiParamTypeClasses
#-}

module Graph.BAC.ProxemyOptim
  where

--import Debug.SimpleReflect
import Data.Eigen.Coeff (coeffSM, coeffDM)
import Data.IntMap (IntMap)
import Data.Maybe (isJust)
import Data.Proxy (Proxy(Proxy))
import Data.Reflection
import Eigen.Internal (CTriplet(..), Elem(..), toC, fromC, C(..), natToInt, Row(..), Col(..))
import Eigen.Matrix (sum)
import Eigen.SparseMatrix (SparseMatrix, SparseMatrixXd, (!), toMatrix)
import GHC.TypeLits (KnownNat, Nat, SomeNat(SomeNat), type(+), natVal, sameNat, someNatVal)
import Graph.FGL
import Graph.Types
import Prelude (String, readLn)
import Protolude hiding (sum, natVal)
import qualified Eigen.Matrix as DenseMatrix
import qualified Data.Graph.Inductive              as DGI
import qualified Data.Graph.Inductive.PatriciaTree as DGIP
import qualified Data.List                         as List
import qualified Data.IntMap                       as Dict
import qualified Data.Vector                       as V
import qualified Data.Vector.Storable              as VS
import qualified Eigen.Matrix                      as DMatrix
import qualified Eigen.SparseMatrix                as SMatrix
import qualified Prelude                           as Prelude
import qualified Data.Set                          as Set


----------------------------------------------------------------


clusteringOptim :: forall n a b. (KnownNat n, Show a, Show b)
                => Length
                -> Similarity
                -> FiniteGraph n a b
                -> Clustering Node
clusteringOptim l s fg@(FiniteGraph g) = -- traceShow ("mc", mc) $
  make_clust_part sorted_edges matq
    where
      adj    = adjacent fg False
      tra    = transition adj

      adj'   = adjacent fg True
      tra'    = transition adj'
      sorted_edges = sort_edges (natToInt @n)
                   $ edges_confluence l adj' tra'

      mc    = matconf False adj (proxemie l tra)
      md    = matmod fg

      matq = case s of
        Conf -> SimConf mc -- $ matconf False adj (proxemie l tra)
        Mod  -> SimMod  md -- $ matmod fg

-- | Where main Types are defined as
data Similarity = Conf | Mod
type Length      = Int

-- | A finite Graph is a Graph whose number of Nodes is a KnownNat
data FiniteGraph (n :: Nat) a b = FiniteGraph (Graph a b)

data Clustering a = ClusteringIs { parts :: !(Dict (Set a))
                                 , index :: !(Dict Int)
                                 , score :: !(Double)
                                 , mode  :: !ClusteringMode
                                 }
                  | Clusterings { strict :: !(StrictClustering   a)
                                , over   :: !(OverlapsClustering a)
                                }
  deriving (Show, Eq)

data ClusteringMode = Part | Over | Both
  deriving (Show, Eq)

type StrictClustering   a = Clustering a
type OverlapsClustering a = Clustering a

type IsReflexive = Bool
type NeighborsFilter a b = DGI.Gr a b -> Node -> [Node]
type RmEdge      = Bool
---------------------------------------------------------------
-- Data Structure
type VectorS          n = SparseMatrix 1 n Double
type VectorD          n = DenseMatrix.Matrix 1 n Double

type AdjacencyMatrix  n = MatrixS n
type TransitionMatrix n = MatrixS n
type ProxemyMatrix    n = MatrixD n

type ConfluenceMatrix n = MatrixD n
type ModularityMatrix n = MatrixD n

data SimilarityMatrix n = SimConf !(ConfluenceMatrix n)
                        | SimMod  !(ModularityMatrix n)


-----
adjacent :: KnownNat n
         => FiniteGraph n a b
         -> IsReflexive
         -> AdjacencyMatrix n
adjacent (FiniteGraph g) isReflexive =
  SMatrix.fromVector $ VS.fromList $ triplets <> diag
  where
    triplets = [ CTriplet (toC i) (toC j) 1.0
               | i <- nodes g
               , j <- neighbors g i
               , i /= j
               ]

    diag = case isReflexive of
      True -> [ CTriplet (toC n) (toC n) 1.0
             | n <- nodes g
             ]
      False -> []


transition :: KnownNat         n
           => AdjacencyMatrix  n
           -> TransitionMatrix n
transition m = SMatrix.imap (\i j v -> v * (VS.!) s i) m
  where
    s = sumWith Colonne (\s -> 1 / s) m


proxemie :: KnownNat         n
         => Length
         -> TransitionMatrix n
         -> ProxemyMatrix    n
proxemie l tm = case l <= 1 of
  True  -> toMatrix tm
  False -> toMatrix $ foldl' (\m' _-> SMatrix.mul m' tm) tm [2 .. (l :: Int)]


---------------------------------------------------------------
matconf :: forall n. KnownNat n
        => RmEdge
        -> AdjacencyMatrix  n
        -> ProxemyMatrix    n
        -> ConfluenceMatrix n
matconf False a p = symmetry confmat
  where
    -- vcount  = natToInt @n
    degs    = sumWith Colonne identity a
    sumdeg  = VS.sum degs
    confmat = DMatrix.imap (\  x   y   v ->
                            if x < y
                              then let
                                      prox_y_x_length = v
                                      prox_y_x_infini = ((VS.!) degs x) / sumdeg
                                 in 
                                    (prox_y_x_length - prox_y_x_infini)
                                  / (prox_y_x_length + prox_y_x_infini)
                              else 0
                        ) $ DMatrix.transpose p
matconf True _a _p = panic "MatConf True: TODO but not needed for now"



matmod :: forall n a b. KnownNat n => FiniteGraph n a b -> ModularityMatrix n
matmod fg = symmetry $ toMatrix modmat
  where
    n' = natToInt @n
    a  = adjacent fg False
    sumRows = sumWith Ligne   identity a
    sumCols = sumWith Colonne identity a
    ecount = sum $ toMatrix a
    modmat = SMatrix.imap (\x y v -> 
                                if x < y
                                then v - ((VS.!) sumRows x * (VS.!) sumCols y) / (2 * ecount)
                                else 0
                          ) a

---------------------------------------------------------------
type UnsortedEdges = [(Node, Node, Double)]
type SortedEdges   = [(Node, Node, Double)]

edges_confluence :: forall           n a b
                  . KnownNat         n
                 => Length
                 -> AdjacencyMatrix  n
                 -> TransitionMatrix n
                 -> UnsortedEdges
edges_confluence l am tm = -- traceShow ("degs", degs) $
  SMatrix.toList matconf'
    where
      vcount   = natToInt @n
      degs     = sumWith Colonne identity am
      sumdeg   = VS.sum degs
      matconf' = SMatrix.imap
                 (\x y _ ->
                   if x < y
                      then
                        let
                            deg_x = (VS.!) degs x - 1
                            deg_y = (VS.!) degs y - 1

                            tm' = SMatrix.imap (\i j v -> if (i == x && j == y) || ( i == y && j == x)
                                                            then 0
                                                            else
                                                              if i == x && i /= y
                                                                then (coeffSM i j am) / deg_x
                                                                else
                                                                   if i == y && i /= x
                                                                     then (coeffSM i j am) / deg_y
                                                                     else v
                                               ) tm

                            v  = fromList (Proxy @n) [(fromIntegral y,1)]
                            v' = doProx l v tm'

                            prox_y_x_length = coeffSM 0 x v'
                            prox_y_x_infini = deg_x / (sumdeg - 2)

                        in -- traceShow ("ec",x,y,deg_x,sumdeg,prox_y_x_infini) $
                             (prox_y_x_length - prox_y_x_infini)
                           / (prox_y_x_length + prox_y_x_infini)

                      else 0
                 ) am

      doProx
        :: forall           n
         . KnownNat         n
        => Length
        -> VectorS     n
        -> TransitionMatrix n
        -> VectorS     n
      doProx l v' tm'' = case l <= 1 of
          True  -> panic "doProx"
          False -> foldl' (\v'' _-> SMatrix.mul v'' tm'') v' [1 .. (l :: Int) ]


-- | TODO optimization
sort_edges :: Int
           -> UnsortedEdges
           -> SortedEdges
sort_edges n =  List.concat
             . (map (List.sortOn (\(x,y,_) -> x * n + y)))
             . (List.groupBy (\x y -> third x == third y))
             . List.reverse
             . (List.filter (\(x,y,_) -> x < y))
             . (List.sortOn third)
  where
    third
      :: forall a b c
       . (a,b,c) -> c
    third (_,_,c) = c


---------------------------------------------------------------
updateClustering :: Clustering Node
                 -> (Int -> Int -> Double)
                 -> Int -> Int -> Clustering Node
updateClustering c@(ClusteringIs parts idx currentScore _) f x y =
  let modX = fromMaybe 0 $ Dict.lookup x idx
      modY = fromMaybe 0 $ Dict.lookup y idx
   in case x == y || modX == modY of
    True  -> c -- do case x' or y' are Nothing 
    False -> let c' = updateWith c f (x,modX) (y,modY)
              in case score c' >= currentScore of
                   True  -> c'
                   False -> c

updateWith :: Clustering Node
           -> (Int -> Int -> Double)
           -> (Int,Int)
           -> (Int,Int)
           -> Clustering Node
updateWith c@(ClusteringIs parts idx _ _) f (x,modX) (y,modY) = -- traceShow ("score", x,y,score') $
  ClusteringIs parts' idx' score' Part
    where
      parts' = Dict.filter (not . Set.null)
             $ Dict.alter (add y) modX
             $ Dict.alter (del y) modY parts

      add y Nothing   = Just (Set.singleton y)
      add y (Just y') = Just (Set.insert y y')

      del y Nothing   = Nothing
      del y (Just y') = Just (Set.delete y y')

      idx'         = Dict.alter (alter modX) y idx
      alter my _   = Just my

      px = Dict.elems parts'

      score' = Prelude.sum $ List.concat
                           $ map (\s -> [ if x'' < y'' then f x'' y'' else 0 -- to Validate with BG
                                        | x'' <- Set.toList s
                                        , y'' <- Set.toList s
                                        ]
                                  ) px

---------------------------------------------------------------
make_clust_part :: forall n
                 . KnownNat n
                => SortedEdges
                -> SimilarityMatrix n
                -> Clustering Node
make_clust_part se sm = foldl' (\c (e1,e2,_)
                                 -> updateClustering c
                                    (\x y -> 2 * (coeffDM x y sm'))
                                    e1 e2
                               ) (ClusteringIs parts idx 0 Part) se
  where
    ns    = [0 .. natToInt @n - 1 ]
    parts = Dict.fromAscList (List.zip ns $ map Set.singleton ns)
    idx   = Dict.fromAscList (List.zip ns ns)
    sm'   = case sm of
      SimConf cm -> cm
      SimMod  mm -> mm

------------------------------
make_clust_over :: KnownNat n
                => SimilarityMatrix n
                -> StrictClustering   a
                -> OverlapsClustering a
make_clust_over = Prelude.undefined

---------------------------------------------------------------
-- | Some Utils

-- | The Main Trick: Graph must be finite, then n can be known 1 time at least
--  The same n is needed the size of the Graph and Matrixes and Vectors
--  Proof by compilation of the compatibilities of the calculations
--  Then Proof by tests is less needed at this specific step of the flow.
buildFiniteGraph :: Proxy n -> Graph a b -> FiniteGraph n a b
buildFiniteGraph _ g = FiniteGraph g

-- | To execute a function on finite Graph
-- the KnownNat n is computed one time only
withG :: (Show a, Show b)
      => Graph a b
      -> (forall n. KnownNat n => FiniteGraph n a b -> r)
      -> r
withG g f = reifyNat (fromIntegral $ length $ nodes g)
          $ \n -> f (buildFiniteGraph n g)

{-
random :: forall n. Proxy n -> IO (MatrixD n)
random _ = DMatrix.random

withM :: Integer
      -> (forall n. KnownNat n => MatrixD n -> r)
      -> r
withM n f = reifyNat (fromIntegral n)
          $ \n -> do
              m <- random n
              pure $ f m
-}


---------------------------
-- Some Sugar to be generalized: same functions for either Dense or
-- sparse Matrix can be unified thanks to the Type instances.
class Symmetry a where
  symmetry :: a -> a

instance KnownNat n => Symmetry (MatrixD n) where
  symmetry = symmetryD

instance KnownNat n => Symmetry (MatrixS n) where
  symmetry = symmetryS

symmetryD :: KnownNat n => MatrixD n -> MatrixD n
symmetryD m = DMatrix.imap (\x y v -> if x < y then v else coeffDM y x m) m

symmetryS :: KnownNat n => MatrixS n -> MatrixS n
symmetryS m = SMatrix.imap (\x y v -> if x < y then v else coeffSM y x m) m

---------------------------

class KnownNat n => Vector n a where
  fromList :: KnownNat n => Proxy n -> [(Integer,Double)] -> a

instance KnownNat n => Vector n (VectorS n) where
  fromList = vectorFromListS

instance KnownNat n => Vector n (VectorD n) where
  fromList = vectorFromListD

vectorFromListS :: KnownNat n => Proxy n -> [(Integer, Double)] -> VectorS n
vectorFromListS pn ns
   | all (\(i, v) -> i >= 0 && i < n) ns = SMatrix.fromList $ map (\(i,v) -> (0, fromInteger i,v)) ns
   | otherwise                           = panic "error vectorFromListS"
     where
       n = natVal pn

vectorFromListD :: KnownNat n => Proxy n -> [(Integer, Double)] -> VectorD n
vectorFromListD pn ns = toMatrix $ vectorFromListS pn ns

------------------------------

data Direction = Ligne | Colonne

sumWith :: ( Elem a
           , Elem t
           , KnownNat n
           ) => Direction -> (t -> a) -> SparseMatrix n n t -> VS.Vector a
sumWith dir f m = VS.fromList
       $ map (\v -> f v)
       $ case dir of
           Colonne -> somme $ SMatrix.getCols m
           Ligne   -> somme $ SMatrix.getRows m
          where
            somme m' = map (sum . SMatrix.toMatrix) m'

--------------------------------------------

